import { Lock } from "@hibas123/utils";

const ALPHABET = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';

const ID_LENGTH = 8;

const generate = () => {
    var rtn = '';
    for (var i = 0; i < ID_LENGTH; i++) {
        rtn += ALPHABET.charAt(Math.floor(Math.random() * ALPHABET.length));
    }
    return rtn;
}

const pathToSegments = (path: IPath) => typeof path === "string" ? path.split("/").filter(e => e !== "") : path;


export type IPath = string | string[];

export interface IDBOperations {
    collection(path: IPath): Collection;
    keys(path: IPath): Promise<string[]>;
    get<T>(path: IPath): Promise<T>;
    set(path: IPath, value: any): Promise<void>;
    push(path: IPath, value: any): Promise<string>;
    subscribe<T>(path: IPath, type: "push" | "set", ondata: (data: T) => void): Promise<{ unsubscribe: () => void }>;
}

export class Collection implements IDBOperations {
    constructor(private db: Database, private prefix: string[]) { }
    collection(path: string | string[]) {
        return new Collection(this.db, [...this.prefix, ...pathToSegments(path)]);
    }

    keys(path: string | string[]) {
        return this.db.keys([...this.prefix, ...pathToSegments(path)])
    }

    get<T = any>(path: string | string[]): Promise<T> {
        return this.db.get([...this.prefix, ...pathToSegments(path)])
    }

    set(path: string | string[], data: any) {
        return this.db.set([...this.prefix, ...pathToSegments(path)], data)
    }

    push(path: string | string[], data: any): Promise<string> {
        return this.db.push([...this.prefix, ...pathToSegments(path)], data)
    }

    subscribe<T>(path: string | string[], type: "push" | "set", ondata: (data: T) => void) {
        return this.db.subscribe([...this.prefix, ...pathToSegments(path)], type, ondata);
    }
}

// TODO: Request Timeout
export default class Database implements IDBOperations {
    static WebSocket: typeof WebSocket;

    private initLock = new Lock();
    private socket: WebSocket;
    private requests = new Map<string, (err: Error, data: any) => void>();
    private eventhandler = new Map<string, (data: any) => void>();

    constructor(url: string, database: string, accesskey?: string, authkey?: string, rootkey?: string) {
        const u = new URL(url);
        if (authkey)
            u.searchParams.set("authkey", authkey);
        if (rootkey)
            u.searchParams.set("rootkey", rootkey);
        if (accesskey)
            u.searchParams.set("accesskey", accesskey);

        u.searchParams.set("database", database);
        this.initLock.getLock().then(lock => {
            this.socket = new Database.WebSocket(u.href);

            this.socket.onopen = () => lock.release();
            this.socket.onerror = err => { console.error(err); lock.release() };
            this.socket.onmessage = msg => {
                let message: { ns: string, data: any } = JSON.parse(msg.data as string);
                switch (message.ns) {
                    case "message":
                        this.handleResponse(message.data);
                        break;
                    case "event":
                        this.handleEvent(message.data)
                        break;
                    case "error_msg":
                        // TODO: Some kind of error handling etc.
                        console.error("Server error:", message.data);
                        break;
                }
            }
        })
    }

    private handleResponse({ id, error, data }: { id: string, error: boolean, data: string }) {
        let req = this.requests.get(id);
        if (req) {
            if (error)
                req(new Error(data), undefined);
            else
                req(undefined, data);
        } else {
            console.error("Invalid response", id);
        }
    }

    private handleEvent({ id, data }: { id: string, data: any }) {
        let eh = this.eventhandler.get(id);
        if (eh)
            eh(data);
    }

    private async queryRequest(type: "get" | "set" | "push" | "keys" | "subscribe" | "unsubscribe", path: IPath, data?: any, id?: string) {
        await this.initLock.getLock().then(l => l.release());
        path = pathToSegments(path)
        if (!id)
            id = generate();

        return new Promise<any>((yes, no) => {
            this.requests.set(id, (err, data) => {
                err ? no(err) : yes(data)
                this.requests.delete(id);
            });
            this.socket.send(JSON.stringify({ ns: "query", data: { id, type, path, data } }))
        })
    }

    collection(path: IPath) {
        return new Collection(this, pathToSegments(path));
    }

    keys(path: IPath) {
        return this.queryRequest("keys", path);
    }

    get<T = any>(path: IPath): Promise<T> {
        return this.queryRequest("get", path);
    }

    set(path: IPath, data: any) {
        return this.queryRequest("set", path, data);
    }

    push(path: IPath, data: any): Promise<string> {
        return this.queryRequest("push", path, data);
    }

    async subscribe(path: IPath, type: "push" | "set", ondata: (data: any) => void) {
        const id = generate();
        const subscriptionID = await this.queryRequest("subscribe", path, type, id);
        this.eventhandler.set(subscriptionID, ondata);
        return {
            unsubscribe: () => {
                this.queryRequest("unsubscribe", path, undefined, id);
                this.eventhandler.delete(subscriptionID);
            }
        }
    }
}